#pragma once

#include <cstdint>
#include "FixedPoint/Fixed.h"
#include "AWSynthSource.h"

class DubSiren
{
    private:
        static constexpr unsigned Q = 23;
        
        static constexpr float R_x(float mod) { return mod < 1.0 ? (3300*50000*mod/(3300 + 50000*mod)) : 3300; }
        static constexpr float U_x0(float mod) { return mod < 1.0 ? (V_cv*50000*mod/(3300 + 50000*mod)) : V_cv; }
        static constexpr float U_x1(float mod) { return mod < 1.0 ? (V_cv + 3300*(V_cc - V_cv)/(3300 + 50000*mod)) : V_cv; }
        
        struct State
        {
            std::uint32_t t0 = 0;
            
             // Output state of the two 555 ICs
            bool ic555_a = false;
            bool ic555_b = false;
            
            // Capacitor volatages
            FixedPoint::Fixed<Q> voltage_c1 = FixedPoint::Fixed<Q>(V_cc);
            FixedPoint::Fixed<Q> voltage_c2 = FixedPoint::Fixed<Q>(V_cc);
        };
    
    public:
        constexpr DubSiren(): _lowpass(0.154) {
            setSpeedAndModulation(0.7, 0.5);
            setFrequency(0.5);
            setPhase(0.5);
        }
        
        constexpr void enableEcho(bool enabled=true) { _echo_enabled = enabled; }
        
        constexpr void enableReverb(bool enabled=true) { _reverb_enabled = enabled; }
        
        constexpr void setSpeedAndModulation(float speed, float mod) {
            constexpr float C1 = 0.000047;
            
            speed = speed < 1 ? (1 - speed) : 0;
            mod = mod < 1 ? (1 - mod) : 0;
            
            _a0 = FixedPoint::Fixed<Q>(U_x0(mod)/(50000 + R_x(mod))*_DELTA_T/C1);
            _b0 = FixedPoint::Fixed<Q>((560 + 50000*speed + 50000 + R_x(mod))/((50000 + R_x(mod))*(560 + 50000*speed))*_DELTA_T/C1);
            _c0 = FixedPoint::Fixed<Q>(U_x0(mod)*50000/(50000 + R_x(mod)));
            
            _a1 = FixedPoint::Fixed<Q>((V_cc/(1120 + 50000*speed) + U_x1(mod)/(50000 + R_x(mod)))*_DELTA_T/C1);
            _b1 = FixedPoint::Fixed<Q>((50000 + R_x(mod) + 1120 + 50000*speed)/((1120 + 50000*speed)*(50000 + R_x(mod)))*_DELTA_T/C1);
            _c1 = FixedPoint::Fixed<Q>(U_x1(mod)*50000/(50000 + R_x(mod)));
            
            _d = FixedPoint::Fixed<Q>(50000/(50000 + R_x(mod)));
        }
        
        constexpr void setFrequency(float freq) {
            constexpr float C2 = 0.00000015;
            
            freq = freq < 1 ? (1 - freq) : 0;
            
            _j0 = FixedPoint::Fixed<Q>(-1.0/(2200 + 50000*freq)*_DELTA_T/C2);
            _i1 = FixedPoint::Fixed<Q>(V_cc/(4400 + 50000*freq)*_DELTA_T/C2);
            _j1 = FixedPoint::Fixed<Q>(1.0/(4400 + 50000*freq)*_DELTA_T/C2);
        }
        
        constexpr void setPhase(float phase) {
            phase = phase < 1 ? (1 - phase) : 0;
            _phase = FixedPoint::Fixed<Q>(phase);
        }
        
        constexpr void reset() {
            for(std::uint32_t idx = 0; idx < 4; ++idx) {
                State& state = _sirens[idx];
                state.t0 = -16*256;
                state.ic555_a = false;
                state.ic555_b = false;
                state.voltage_c1 = FixedPoint::Fixed<Q>(V_cc);
                state.voltage_c2 = FixedPoint::Fixed<Q>(V_cc);
            }
            _sirens[0].t0 = 0;
            
            _echo_index = 0;
            _next_t0 = ECHO_DELAY;
        }
        
        std::int32_t callback(std::uint32_t t, std::uint32_t p) {
            using AWSynth = Audio::AWSynthSource;
            
            std::int32_t siren_out = 0;
            if(_echo_enabled) {
                if(t >= _next_t0) {
                    _echo_index = (_echo_index + 1)&3;
                    State& state = _sirens[_echo_index];
                    state.t0 = _next_t0;
                    state.ic555_a = false;
                    state.ic555_b = false;
                    state.voltage_c1 = FixedPoint::Fixed<Q>(V_cc);
                    state.voltage_c2 = FixedPoint::Fixed<Q>(V_cc);
                    
                    do {
                        _next_t0 += ECHO_DELAY;
                    } while(_next_t0 <= t);
                }
                
                siren_out = tick(_sirens[0], (1<<14) - AWSynth::ramp<12*256>(t - _sirens[0].t0));
                siren_out += tick(_sirens[1], (1<<14) - AWSynth::ramp<12*256>(t - _sirens[1].t0));
                siren_out += tick(_sirens[2], (1<<14) - AWSynth::ramp<12*256>(t - _sirens[2].t0));
                siren_out += tick(_sirens[3], (1<<14) - AWSynth::ramp<12*256>(t - _sirens[3].t0));
            }
            else {
                siren_out = tick(_sirens[0], 1<<15);
            }
            
            if(_reverb_enabled) {
                // Gain is 0.001^(tau/RVT) where tau is delay time and RVT is the desired reverb time
                constexpr std::int32_t G1 = 0.980*(1<<10);
                constexpr std::int32_t G2 = 0.975*(1<<10);
                constexpr std::int32_t G3 = 0.972*(1<<10);
                constexpr std::int32_t G4 = 0.970*(1<<10);
                
                siren_out = (siren_out + (1<<3))>>4;  // Compensate for amplification by the filters
                
                // Four tap comb filter
                std::int32_t comb_out = (G1*_comb.peek(946) + G2*_comb.peek(1186) + G3*_comb.peek(1318) + G4*_comb.peek(1398) + (1<<11))>>(10 + 2);
                comb_out = siren_out + _lowpass.propagate(comb_out);
                _comb.propagate(comb_out);
                
                // First all-pass reverb module
                constexpr std::int32_t G_a = 0.700*(1<<10);
                std::int32_t allpass1_out = _allpass1.peek(156) - ((G_a*comb_out + (1<<9))>>10);
                _allpass1.propagate(comb_out + ((G_a*allpass1_out + (1<<9))>>10));
                
                // Second all-pass reverb module
                constexpr std::int32_t G_b = 0.700*(1<<10);
                std::int32_t allpass2_out = _allpass2.peek(52) - ((G_b*allpass1_out + (1<<9))>>10);
                _allpass2.propagate(comb_out + ((G_b*allpass2_out + (1<<9))>>10));
                
                siren_out = allpass2_out;
            }
            
            return (siren_out + (1<<7))>>8; // Scale output to 8-bits
        }
        
        std::int32_t tick(State& state, std::int32_t amplitude) {
            FixedPoint::Fixed<Q> voltage_cv_b;
            
            // First 555
            state.ic555_a = ic555(state.ic555_a, state.voltage_c1, state.voltage_c1);
            if(state.ic555_a) {
                voltage_cv_b = state.voltage_c1 + _phase*(_c1 - state.voltage_c1*_d);
                state.voltage_c1 += _a1 - state.voltage_c1*_b1;
            }
            else {
                voltage_cv_b = state.voltage_c1 + _phase*(_c0 - state.voltage_c1*_d);
                state.voltage_c1 += _a0 - state.voltage_c1*_b0;
            }
            
            // Second 555
            state.ic555_b = ic555(state.ic555_b, state.voltage_c2, state.voltage_c2, voltage_cv_b);
            if(state.ic555_b) {
                state.voltage_c2 += _i1 - state.voltage_c2*_j1;
            }
            else {
                state.voltage_c2 += state.voltage_c2*_j0;
            }
            
            return state.ic555_b ? amplitude : -amplitude;
        }
    
    private:
        template<typename T, unsigned MAX_DELAY>
        struct MultitapDelay
        {
            constexpr T peek(std::uint32_t offset=0) {
                return _buffer[_head + offset];
            }
            
            inline void propagate(T value) {
                _head = _head > 0 ? (_head - 1) : (MAX_DELAY - 1);
                
                // Write the value to two locations to enable faster multitap read
                _buffer[_head] = value;
                _buffer[_head + MAX_DELAY] = value;
            }
            
            T _buffer[2*MAX_DELAY] = {};
            std::uint32_t _head = 0;
        };
        
        template<typename T>
        struct LowPass
        {
            constexpr LowPass(float a0): _coeff_x((a0/2)*(1<<10)), _coeff_y((1 - a0)*(1<<10)) {}
            
            inline T propagate(const T& xn) {
                std::int32_t yn = (std::int64_t(_coeff_y)*_yn1 + std::int64_t(_coeff_x)*(xn + _xn1) + (1<<9))>>10;
                _xn1 = xn;
                _yn1 = yn;
                return yn;
            }
            
            const std::int32_t _coeff_x;
            const std::int32_t _coeff_y;
            T _xn1 = 0;
            T _yn1 = 0;
        };
        
        static constexpr float _DELTA_T = 1.0/PROJ_AUD_FREQ;
        
        static constexpr float V_cc = 1.0;
        static constexpr float V_cv = V_cc*2/3;
        
        static constexpr std::uint32_t ECHO_DELAY = 4*256;
        
        FixedPoint::Fixed<Q> _a0 = FixedPoint::Fixed<Q>(0);
        FixedPoint::Fixed<Q> _b0 = FixedPoint::Fixed<Q>(0);
        FixedPoint::Fixed<Q> _c0 = FixedPoint::Fixed<Q>(0);
        
        FixedPoint::Fixed<Q> _a1 = FixedPoint::Fixed<Q>(0);
        FixedPoint::Fixed<Q> _b1 = FixedPoint::Fixed<Q>(0);
        FixedPoint::Fixed<Q> _c1 = FixedPoint::Fixed<Q>(0);
        
        FixedPoint::Fixed<Q> _d = FixedPoint::Fixed<Q>(0);
        
        FixedPoint::Fixed<Q> _j0 = FixedPoint::Fixed<Q>(0);
        FixedPoint::Fixed<Q> _i1 = FixedPoint::Fixed<Q>(0);
        FixedPoint::Fixed<Q> _j1 = FixedPoint::Fixed<Q>(0);
        
        FixedPoint::Fixed<Q> _phase = FixedPoint::Fixed<Q>(0);
        
        bool _echo_enabled = false;
        bool _reverb_enabled = false;
        
        State _sirens[4] = {};
        std::uint32_t _echo_index = 0;
        std::uint32_t _next_t0 = ECHO_DELAY;
        
        MultitapDelay<std::int32_t, 1600> _comb;
        LowPass<std::int32_t> _lowpass;
        MultitapDelay<std::int32_t, 157> _allpass1;
        MultitapDelay<std::int32_t, 53> _allpass2;
        
        bool ic555(bool prev_state, const FixedPoint::Fixed<Q>& trigger, const FixedPoint::Fixed<Q>& threshold, const FixedPoint::Fixed<Q>& control_voltage=FixedPoint::Fixed<Q>(V_cc*2/3)) {
            if(trigger < FixedPoint::Fixed<Q>(V_cc*1/3)) {
                return true;
            }
            else if(threshold > control_voltage) {
                return false;
            }
            return prev_state;
        }
};
